#define N           5               /* number of philosophers */
#define LEFT        ( i - 1 ) % N   /* number of i's left neighbor */
#define RIGHT       ( i + 1 ) % N   /* number of i's right neighbor */
#define THINKING    0               /* philosopher is thinking */
#define HUNGRY      1               /* philosopher is hungry */
#define EATING      2               /* philosopher is eating */

typdef int semaphore;   /* semaphores are a special kind of int */

int state[N];           /* array to keep track of everyone's state */

semaphore mutex = 1;    /* mutual exclusion for critical regions */
semaphore s[N];         /* one semaphore per philosopher */

/* i: philosopher number, from 0 to N-1 */
void philosopher( int i )
{
    while( TRUE )       /* repeat forever */
    {
        think();        /* philosopher is thinking */
        take_forks();   /* acquire two forks or block */
        eat();          /* yum-yum, spaghetti */
        put_forks();    /* put both forks back on table */
    }
}

/* i: philosopher number, from 0 to N-1 */
void take_forks( int i )
{
    down( &mutex );         /* enter critical region */
    state[i] = HUNGRY;      /* record that philosopher i is hungry */
    test( i );              /* try to acquire 2 forks */
    up( &mutex );           /* exit critical region */
    down( &s[i] );          /* block if forks were not aquired */
}

/* i: philosopher number, from 0 to N-1 */
void put_forks( int i )
{
    down( &mutex );         /* enter critical region */
    state[i] = THINKING;    /* philosopher has finished eating */
    test( LEFT  );          /* see if left  neighbor can now eat */
    test( RIGHT );          /* see if right neighbor can now eat */
    up( &mutex );           /* exit critical region */
}

void test( int i )
{
    if( state[i]     == HUNGRY &&
        state[LEFT]  != EATING &&
        state[RIGHT] != EATING )
    {
        state[i] = EATING;
        up( &s[i] );
    }
}